from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers

from wbhuman_resources.models import (
    DayOff,
    DefaultDailyPeriod,
    EmployeeWeeklyOffPeriods,
)
from wbhuman_resources.models.calendars import DayOffCalendar


class DayOffCalendarRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = DayOffCalendar
        fields = ("id", "title")


class DefaultDailyPeriodRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = DefaultDailyPeriod
        fields = ("id", "lower_time", "upper_time", "title", "total_hours")


class EmployeeWeeklyOffPeriodsRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = EmployeeWeeklyOffPeriods
        fields = ("id", "computed_str")


class DayOffRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = DayOff
        fields = ("id", "title", "date")


class DayOffCalendarModelSerializer(wb_serializers.ModelSerializer):
    timezone = wb_serializers.TimeZoneField()

    @wb_serializers.register_resource()
    def additional_resources(self, instance, request, user):
        return {
            "days_off": reverse(
                "wbhuman_resources:calendar-dayoff-list",
                args=[instance.id],
                request=request,
            ),
            "default_periods": reverse(
                "wbhuman_resources:calendar-defaultperiod-list",
                args=[instance.id],
                request=request,
            ),
        }

    class Meta:
        model = DayOffCalendar
        fields = ("id", "title", "resource", "timezone", "_additional_resources")


class DefaultDailyPeriodModelSerializer(wb_serializers.ModelSerializer):
    _calendar = DayOffCalendarRepresentationSerializer(source="calendar")

    class Meta:
        model = DefaultDailyPeriod
        fields = ("id", "lower_time", "upper_time", "timespan", "title", "total_hours", "calendar", "_calendar")


class DayOffModelSerializer(wb_serializers.ModelSerializer):
    _calendar = DayOffCalendarRepresentationSerializer(source="calendar")

    class Meta:
        model = DayOff
        fields = ("id", "title", "date", "count_as_holiday", "calendar", "_calendar")



import zoneinfo
from datetime import datetime, time, timedelta
from importlib import import_module

import pytest
from django.db.models import Q
from django.utils.timezone import make_aware
from faker import Faker
from psycopg.types.range import TimestamptzRange

from wbhuman_resources.models.calendars import (
    DayOff,
    InvalidDayOffCalendarResourceError,
)
from wbhuman_resources.models.employee import EmployeeHumanResource

fake = Faker()


@pytest.mark.django_db
class TestDayOffCalendar:
    def test_get_period_start_choices(self, day_off_calendar):
        assert day_off_calendar.get_period_start_choices() == ["09:00:00", "14:00:00"]

    def test_get_period_end_choices(self, day_off_calendar):
        assert day_off_calendar.get_period_end_choices() == ["13:00:00", "18:00:00"]

    @pytest.mark.parametrize("val_date", [(fake.date_this_decade())])
    def test_create_public_holidays_sanitize_resource(self, day_off_calendar_factory, val_date):
        with pytest.raises(InvalidDayOffCalendarResourceError):
            calendar = day_off_calendar_factory.create(resource="eurasia.China")
            calendar.create_public_holidays(val_date.year)
        with pytest.raises(InvalidDayOffCalendarResourceError):
            calendar = day_off_calendar_factory.create(resource="europe.Listenbourg")
            calendar.create_public_holidays(val_date.year)

    @pytest.mark.parametrize(
        "val_date,continent,region",
        [
            (fake.date_this_decade(), "europe", "Berlin"),
            (fake.date_this_decade(), "europe", "Switzerland"),
        ],
    )
    def test_create_public_holidays_valid_resource(self, day_off_calendar_factory, val_date, continent, region):
        calendar = day_off_calendar_factory.create(resource=f"{continent}.{region}")
        calendar.create_public_holidays(val_date.year)
        workalendar = import_module(f"workalendar.{continent}")
        cal = getattr(workalendar, region)()
        for _d, _ in cal.holidays(val_date.year):
            assert DayOff.objects.filter(
                date=_d,
                calendar=calendar,
            )

    def test_get_day_off_per_employee_df(self, day_off_calendar, employee_human_resource_factory, day_off_factory):
        employee1 = employee_human_resource_factory.create()
        employee2 = employee_human_resource_factory.create()
        employee_human_resource_factory.create()
        base_day_off = day_off_factory.create()
        period1 = day_off_calendar.default_periods.first()
        period2 = day_off_calendar.default_periods.last()
        day_off_factory.create(date=base_day_off.date - timedelta(days=1))  # create left_outside_day_off
        day_off_factory.create(date=base_day_off.date + timedelta(days=1))  # create right_outside_day_off
        res = (
            day_off_calendar.get_day_off_per_employee_df(
                base_day_off.date,
                base_day_off.date,
                EmployeeHumanResource.objects.filter(Q(id=employee1.id) | Q(id=employee2.id)),
            )
            .set_index(["employee", "period"])
            .to_dict("index")
        )
        assert res == {
            (employee1.id, period1.id): {"date": base_day_off.date, "type": "Holiday", "status": "APPROVED"},
            (employee1.id, period2.id): {"date": base_day_off.date, "type": "Holiday", "status": "APPROVED"},
            (employee2.id, period1.id): {"date": base_day_off.date, "type": "Holiday", "status": "APPROVED"},
            (employee2.id, period2.id): {"date": base_day_off.date, "type": "Holiday", "status": "APPROVED"},
        }

    @pytest.mark.parametrize("h1,h2", [(fake.pyint(min_value=1), fake.pyint(min_value=1))])
    def test_get_daily_hours(self, day_off_calendar_without_period, default_daily_period_factory, h1, h2):
        default_daily_period_factory.create(
            calendar=day_off_calendar_without_period,
            total_hours=h1,
            lower_time=time(9, 0, 0),
            upper_time=time(13, 0, 0),
        )
        default_daily_period_factory.create(
            calendar=day_off_calendar_without_period,
            total_hours=h2,
            lower_time=time(14, 0, 0),
            upper_time=time(18, 0, 0),
        )
        assert day_off_calendar_without_period.get_daily_hours() == h1 + h2

    @pytest.mark.parametrize(
        "ranges, hour_start, expected_res",
        [
            (
                [(time(9, 0, 0), time(13, 0, 0)), (time(14, 0, 0), time(18, 0, 0))],
                time(0, 0),
                [(time(0, 0), time(8, 59)), (time(13, 1), time(13, 59)), (time(18, 1), time(23, 59))],
            ),
            (
                [(time(9, 0, 0), time(13, 0, 0)), (time(14, 0, 0), time(18, 0, 0))],
                time(4, 0),
                [(time(4, 0), time(8, 59)), (time(13, 1), time(13, 59)), (time(18, 1), time(3, 59))],
            ),
            (
                [(time(9, 0, 0), time(13, 0, 0)), (time(14, 0, 0), time(18, 0, 0))],
                time(15, 0),
                [(time(18, 1), time(8, 59)), (time(13, 1), time(13, 59))],
            ),
            (
                [(time(9, 0, 0), time(13, 0, 0)), (time(14, 0, 0), time(18, 0, 0))],
                time(20, 0),
                [(time(20, 0), time(8, 59)), (time(13, 1), time(13, 59)), (time(18, 1), time(19, 59))],
            ),
        ],
    )
    def test_get_unworked_time_range(
        self, day_off_calendar_without_period, default_daily_period_factory, ranges, hour_start, expected_res
    ):
        for lower, upper in ranges:
            default_daily_period_factory.create(
                calendar=day_off_calendar_without_period, lower_time=lower, upper_time=upper
            )

        assert list(day_off_calendar_without_period.get_unworked_time_range(hour_start)) == expected_res

    @pytest.mark.parametrize(
        "unnormalized_lower_datetime, unnormalized_upper_datetime, expected_lower_datetime, expected_upper_datetime",
        [
            (datetime(2023, 1, 1, 8), datetime(2023, 1, 1, 20), datetime(2023, 1, 1, 9), datetime(2023, 1, 1, 18)),
            (datetime(2023, 1, 1, 15), datetime(2023, 1, 1, 23), datetime(2023, 1, 1, 14), datetime(2023, 1, 1, 18)),
            (
                datetime(2023, 1, 1, 2, 3, 2),
                datetime(2023, 1, 1, 23, 5, 6),
                datetime(2023, 1, 1, 9),
                datetime(2023, 1, 1, 18),
            ),
        ],
    )
    def test_normalize_period(
        self,
        day_off_calendar,
        unnormalized_lower_datetime,
        unnormalized_upper_datetime,
        expected_lower_datetime,
        expected_upper_datetime,
    ):
        normalized_period = day_off_calendar.normalize_period(
            TimestamptzRange(
                lower=make_aware(unnormalized_lower_datetime, day_off_calendar.timezone),
                upper=make_aware(unnormalized_upper_datetime, day_off_calendar.timezone),
            )
        )
        assert normalized_period.lower == make_aware(expected_lower_datetime, day_off_calendar.timezone)
        assert normalized_period.upper == make_aware(expected_upper_datetime, day_off_calendar.timezone)


@pytest.mark.django_db
class TestDayOff:
    @pytest.mark.parametrize(
        "timezone_str, lower_time, upper_time",
        [
            ("Pacific/Kwajalein", time(9, 0, 0), time(18, 0, 0)),
            ("Europe/Berlin", time(9, 0, 0), time(18, 0, 0)),
            ("Europe/Berlin", time(15, 0, 0), time(16, 0, 0)),
        ],
    )
    def test_get_timespan(
        self,
        base_day_off_calendar_factory,
        default_daily_period_factory,
        day_off_factory,
        timezone_str,
        lower_time,
        upper_time,
    ):
        timezone = zoneinfo.ZoneInfo(timezone_str)
        calendar = base_day_off_calendar_factory.create(timezone=timezone)
        default_daily_period_factory.create(calendar=calendar, lower_time=lower_time, upper_time=upper_time)
        day_off = day_off_factory.create(calendar=calendar)
        assert day_off.period.lower == datetime.combine(day_off.date, lower_time, tzinfo=timezone).astimezone(
            zoneinfo.ZoneInfo("UTC")
        )
        assert day_off.period.upper == datetime.combine(day_off.date, upper_time, tzinfo=timezone).astimezone(
            zoneinfo.ZoneInfo("UTC")
        )
        #
        # timespan = day_off.get_timespan(as_utc=False)
        # assert timespan.lower == datetime.combine(day_off.date, lower_time, tzinfo=timezone)
        # assert timespan.upper == datetime.combine(day_off.date, upper_time, tzinfo=timezone)


@pytest.mark.django_db
class TestDefaultDailyPeriod:
    @pytest.mark.parametrize("val_date", [fake.date_object()])
    def test_get_lower_datetime(self, default_daily_period, val_date):
        assert default_daily_period.get_lower_datetime(val_date) == datetime.combine(
            val_date, default_daily_period.lower_time, tzinfo=default_daily_period.calendar.timezone
        ).astimezone(zoneinfo.ZoneInfo("UTC"))

    @pytest.mark.parametrize("val_date", [fake.date_object()])
    def test_get_upper_datetime(self, default_daily_period, val_date):
        assert default_daily_period.get_upper_datetime(val_date) == datetime.combine(
            val_date, default_daily_period.upper_time, tzinfo=default_daily_period.calendar.timezone
        ).astimezone(zoneinfo.ZoneInfo("UTC"))



# Generated by Django 4.1.9 on 2023-06-02 06:15

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbhuman_resources", "0001_initial_squashed_squashed_0015_alter_absencerequest_calendaritem_ptr_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="employeehumanresource",
            options={
                "permissions": [("global_manager", "Is a Global Manager?")],
                "verbose_name": "Employee Human Resource",
                "verbose_name_plural": "Employee Human Resources",
            },
        ),
    ]



from django.utils.translation import gettext as _
from wbcore.metadata.configs.titles import TitleViewConfig

from wbhuman_resources.models import Review, ReviewGroup, ReviewQuestionCategory


class ReviewTemplateTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return _("Review Templates")


class ReviewReviewGroupTitleConfig(TitleViewConfig):
    def get_list_title(self):
        if review_group_id := self.view.kwargs.get("review_group_id"):
            review_group = ReviewGroup.objects.get(id=review_group_id)
            return _("Review Group: {review_group}").format(review_group=review_group.name)
        return super().get_list_title()


class ReviewQuestionReviewQuestionCategoryTitleConfig(TitleViewConfig):
    def get_list_title(self):
        if category_id := self.view.kwargs.get("category_id"):
            category = ReviewQuestionCategory.objects.get(id=category_id)
            return _("Review Question Category: {category}").format(category=category.name)
        return super().get_list_title()


class ReviewQuestionReviewTitleConfig(TitleViewConfig):
    def get_list_title(self):
        if review_id := self.view.kwargs.get("review_id"):
            review = Review.objects.get(id=review_id)
            return _("Review: {review}").format(review=str(review))
        return super().get_list_title()


class ReviewAnswerReviewTitleConfig(TitleViewConfig):
    def get_list_title(self):
        if review_id := self.view.kwargs.get("review_id"):
            review = Review.objects.get(id=review_id)
            return _("Review: {review}").format(review=str(review))
        return super().get_list_title()


class ReviewProgressReviewPandasTitleConfig(TitleViewConfig):
    def get_list_title(self):
        if review_id := self.view.kwargs.get("review_id"):
            review = Review.objects.get(id=review_id)
            return _("Progress of the review : {review}").format(review=str(review))
        return _("Progress of the review")


class ReviewAnswerReviewPandasTitleConfig(TitleViewConfig):
    def get_list_title(self):
        if review_id := self.view.kwargs.get("review_id"):
            review = Review.objects.get(id=review_id)
            return _("Evaluation of the review : {review}").format(review=str(review))
        return _("Evaluation of the review")


class ReviewProgressPandasTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return _("Progress of Reviews")



from django.db.models import F, OuterRef, Subquery
from rest_framework.decorators import action
from rest_framework.response import Response
from wbcore import viewsets
from wbcore.contrib.directory.models import Person
from wbcore.utils.date import current_month_date_end

from wbhuman_resources.filters import (
    EmployeeBalanceFilterSet,
    EmployeeFilterSet,
    PositionFilterSet,
)
from wbhuman_resources.models import (
    EmployeeHumanResource,
    EmployeeWeeklyOffPeriods,
    EmployeeYearBalance,
    Position,
    deactivate_profile_as_task,
)
from wbhuman_resources.serializers import (
    EmployeeBalanceModelSerializer,
    EmployeeHumanResourceRepresentationSerializer,
    EmployeeModelSerializer,
    EmployeeWeeklyOffPeriodsModelSerializer,
    EmployeeYearBalanceModelSerializer,
    EmployeeYearBalanceRepresentationSerializer,
    PositionModelSerializer,
    PositionRepresentationSerializer,
)
from wbhuman_resources.viewsets.buttons import (
    EmployeeButtonConfig,
    YearBalanceEmployeeHumanResourceButtonConfig,
)
from wbhuman_resources.viewsets.display import (
    EmployeeBalanceDisplayConfig,
    EmployeeDisplayConfig,
    PositionDisplayConfig,
    WeeklyOffPeriodEmployeeHumanResourceDisplayConfig,
    YearBalanceEmployeeHumanResourceDisplayConfig,
)
from wbhuman_resources.viewsets.endpoints import (
    EmployeeBalanceEndpointConfig,
    WeeklyOffPeriodEmployeeHumanResourceEndpointConfig,
    YearBalanceEmployeeHumanResourceEndpointConfig,
)
from wbhuman_resources.viewsets.titles import (
    EmployeeBalanceTitleConfig,
    EmployeeTitleConfig,
)

from .mixins import EmployeeViewMixin


class PositionRepresentationViewSet(viewsets.RepresentationViewSet):
    ordering_fields = ordering = ("name",)
    search_fields = ("name",)
    filterset_fields = {"height": ["exact"], "level": ["exact"]}
    queryset = Position.objects.all()
    serializer_class = PositionRepresentationSerializer


class EmployeeHumanResourceRepresentationViewSet(viewsets.RepresentationViewSet):
    ordering_fields = ordering = ("profile__computed_str",)
    search_fields = ("profile__computed_str",)

    queryset = EmployeeHumanResource.active_employees.select_related("profile")
    serializer_class = EmployeeHumanResourceRepresentationSerializer


class EmployeeYearBalanceRepresentationViewSet(viewsets.RepresentationViewSet):
    ordering_fields = ordering = "-year"

    queryset = EmployeeYearBalance.objects.all()
    serializer_class = EmployeeYearBalanceRepresentationSerializer


class PositionModelViewSet(viewsets.ModelViewSet):
    queryset = Position.objects.select_related(
        "parent",
        "manager",
    )
    serializer_class = PositionModelSerializer

    filterset_class = PositionFilterSet
    ordering_fields = ordering = ["name"]
    search_fields = ["name"]

    display_config_class = PositionDisplayConfig


class EmployeeBalanceModelViewSet(EmployeeViewMixin, viewsets.ModelViewSet):
    queryset = EmployeeHumanResource.objects.all()
    serializer_class = EmployeeBalanceModelSerializer

    filterset_class = EmployeeBalanceFilterSet
    ordering_fields = [
        "profile__computed_str",
        "contract_type",
        "is_active",
        "position__name",
        "extra_days_frequency",
        "extra_days_per_period",
        "took_long_vacations",
        "available_vacation_balance_previous_year",
        "available_vacation_balance_current_year",
        "available_vacation_balance_next_year",
    ]
    ordering = ["profile__computed_str"]
    search_fields = ["profile__computed_str"]

    title_config_class = EmployeeBalanceTitleConfig
    display_config_class = EmployeeBalanceDisplayConfig
    endpoint_config_class = EmployeeBalanceEndpointConfig

    def get_queryset(self):
        qs = EmployeeHumanResource.objects.none()
        if self.is_administrator:
            qs = super().get_queryset()
        if employee := getattr(self.request.user.profile, "human_resources", None):
            qs = employee.get_managed_employees()
        return (
            EmployeeHumanResource.annotated_queryset(qs, current_month_date_end())
            .filter(balances__isnull=False)
            .distinct()
        ).select_related(
            "profile",
            "position",
            "calendar",
        )


class EmployeeModelViewSet(EmployeeViewMixin, viewsets.ModelViewSet):
    queryset = EmployeeHumanResource.objects.all()
    serializer_class = EmployeeModelSerializer

    ordering_fields = [
        "profile__computed_str",
        "position__name",
        "top_position_repr",
        "primary_telephone",
        "primary_email",
        "primary_address",
    ]
    ordering = ["profile__computed_str"]
    search_fields = ["profile__computed_str", "primary_telephone", "primary_email", "primary_address"]

    filterset_class = EmployeeFilterSet

    title_config_class = EmployeeTitleConfig
    display_config_class = EmployeeDisplayConfig
    button_config_class = EmployeeButtonConfig

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(
                top_position_repr=F("position__parent__name"),
                position_manager=F("position__manager"),
                primary_email=Subquery(
                    Person.objects.filter(id=OuterRef("profile")).annotate_all().values("primary_email")[:1]
                ),
                primary_address=Subquery(
                    Person.objects.filter(id=OuterRef("profile")).annotate_all().values("primary_address")[:1]
                ),
                primary_telephone=Subquery(
                    Person.objects.filter(id=OuterRef("profile")).annotate_all().values("primary_telephone")[:1]
                ),
            )
            .select_related(
                "direct_manager",
                "position",
                "calendar",
                "profile",
            )
            .prefetch_related("balances", "position__manager")
        )

    @action(detail=True, methods=["PATCH"])
    def deactivate(self, request, pk=None):
        if pk and EmployeeHumanResource.is_administrator(self.request.user):
            deactivate_profile_as_task.delay(request.user.id, pk, request.POST.get("substitute", None))
        return Response({"send": True})


# Employee Subs Viewsets


class WeeklyOffPeriodEmployeeHumanResourceModelViewSet(viewsets.ModelViewSet):
    display_config_class = WeeklyOffPeriodEmployeeHumanResourceDisplayConfig
    endpoint_config_class = WeeklyOffPeriodEmployeeHumanResourceEndpointConfig

    serializer_class = EmployeeWeeklyOffPeriodsModelSerializer
    queryset = EmployeeWeeklyOffPeriods.objects.all()

    def get_queryset(self):
        return super().get_queryset().filter(employee=self.kwargs["employee_id"])


class YearBalanceEmployeeHumanResourceModelViewset(viewsets.ModelViewSet):
    display_config_class = YearBalanceEmployeeHumanResourceDisplayConfig
    endpoint_config_class = YearBalanceEmployeeHumanResourceEndpointConfig
    button_config_class = YearBalanceEmployeeHumanResourceButtonConfig

    ordering = ["-year"]
    serializer_class = EmployeeYearBalanceModelSerializer
    queryset = EmployeeYearBalance.objects.all()

    def get_queryset(self):
        return super().get_queryset().filter(employee=self.kwargs["employee_id"])



import random

import factory
from faker import Faker
from pandas.tseries.offsets import BDay
from psycopg.types.range import TimestamptzRange
from wbcore.contrib.icons import WBIcon

from wbhuman_resources.models import (
    AbsenceRequest,
    AbsenceRequestPeriods,
    AbsenceRequestType,
)

from .calendars import DefaultDailyPeriodFactory
from .employee import EmployeeYearBalanceFactory

fake = Faker()


class AbsenceRequestTypeFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = AbsenceRequestType
        django_get_or_create = ["title"]

    title = factory.Faker("text", max_nb_chars=64)
    icon = factory.Iterator(WBIcon.values)
    is_vacation = factory.Faker("boolean")
    is_timeoff = factory.Faker("boolean")
    is_extensible = factory.Faker("boolean")
    auto_approve = False
    days_in_advance = factory.Faker("pyint", min_value=0, max_value=5)

    @factory.post_generation
    def crossborder_countries(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for country in extracted:
                self.crossborder_countries.add(country)


def _get_random_period(calendar):
    lower = Faker().future_datetime() + BDay(0)
    upper = lower + BDay(random.randint(1, 7))
    return TimestamptzRange(
        lower=lower.to_pydatetime().astimezone(calendar.timezone),
        upper=upper.to_pydatetime().astimezone(calendar.timezone),
    )


class VacationTypeFactory(AbsenceRequestTypeFactory):
    title = "Vacation"
    is_vacation = True
    is_timeoff = True
    is_extensible = False
    auto_approve = False


class TimeOffTypeFactory(AbsenceRequestTypeFactory):
    title = "TimeOff"
    is_vacation = False
    is_timeoff = True
    is_extensible = True
    auto_approve = True


class AbsenceRequestFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = AbsenceRequest

    period = factory.LazyAttribute(lambda o: _get_random_period(o.employee.calendar))
    attachment = factory.django.FileField()
    type = factory.SubFactory("wbhuman_resources.factories.AbsenceRequestTypeFactory")
    employee = factory.SubFactory("wbhuman_resources.factories.EmployeeHumanResourceFactory")
    notes = factory.Faker("text")
    reason = factory.Faker("text")


class VacationRequestFactory(AbsenceRequestFactory):
    type = factory.SubFactory(VacationTypeFactory)
    status = AbsenceRequest.Status.APPROVED


class TimeOffRequestFactory(AbsenceRequestFactory):
    type = factory.SubFactory(TimeOffTypeFactory)
    status = AbsenceRequest.Status.APPROVED


class AbsenceRequestPeriodsFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = AbsenceRequestPeriods
        django_get_or_create = ["employee", "default_period", "date"]

    date = factory.LazyAttribute(lambda o: (fake.date_object() + BDay(0)).date())
    employee = factory.SubFactory("wbhuman_resources.factories.EmployeeHumanResourceFactory")
    default_period = factory.LazyAttribute(lambda o: DefaultDailyPeriodFactory.create(calendar=o.employee.calendar))
    request = factory.LazyAttribute(
        lambda o: AbsenceRequestFactory.create(
            status=AbsenceRequest.Status.APPROVED,
            type=VacationTypeFactory.create(),
            employee=o.employee,
            period=TimestamptzRange(
                lower=o.default_period.get_lower_datetime(o.date), upper=o.default_period.get_upper_datetime(o.date)
            ),
        )
    )
    balance = factory.LazyAttribute(lambda o: EmployeeYearBalanceFactory.create(year=o.date.year, employee=o.employee))



from .absence import (
    AbsenceRequestEmployeeBalanceTitleConfig,
    AbsenceRequestPlannerTitleConfig,
    AbsenceTablePandasTitleConfig,
    AbsenceTypeCountEmployeeTitleConfig,
)
from .employee import EmployeeBalanceTitleConfig, EmployeeTitleConfig
from .kpis import EvaluationGraphTitleConfig, KPIEvaluationPandasTitleConfig
from .review import (
    ReviewAnswerReviewPandasTitleConfig,
    ReviewAnswerReviewTitleConfig,
    ReviewProgressPandasTitleConfig,
    ReviewProgressReviewPandasTitleConfig,
    ReviewQuestionReviewQuestionCategoryTitleConfig,
    ReviewQuestionReviewTitleConfig,
    ReviewReviewGroupTitleConfig,
    ReviewTemplateTitleConfig,
)



import importlib
from datetime import date
from typing import Iterator, Protocol, Type

import numpy as np
import pandas as pd
from django.conf import settings
from django.contrib.postgres.fields import DateRangeField
from django.db import models
from django.db.models.query import QuerySet
from django.utils.translation import gettext_lazy as _
from wbcore.models import WBModel
from wbcore.serializers.serializers import ModelSerializer


class KPIHandler(Protocol):
    """
    This is a protocol to ensure that two methods are present in the implemented handlers.
    A handler makes sure that we have a valid serializer and a valid evaluate method to customize
    KPIs.
    """

    def get_name(self) -> str: ...

    def get_serializer(self) -> Type[ModelSerializer]: ...

    def annotate_parameters(self, queryset: QuerySet["KPI"]) -> QuerySet["KPI"]: ...

    def evaluate(self, kpi: "KPI") -> int: ...


def default_additional_data() -> dict[str, dict | list]:
    return {
        "serializer_data": {},
        "list_data": [],
    }


class KPI(WBModel):
    """
    Key Performance Indicator. Stores 1 KPI for either a person or a group of people. It defines a goal and a method to evaluate this goal.
    Additional parameters are set through a custom serializer which stores its result in a JSON field. The visualizion of it is done by
    passing the json field in the custom serializer. The addition parameters in the json field need to have to keys, one for the raw data that
    is passed into the form serializer and one for a list to show the parameters in human readable form in a list field.
    """

    class Interval(models.TextChoices):
        DAILY = "DAILY", _("Daily")
        WEEKLY = "WEEKLY", _("Weekly")
        MONTHLY = "MONTHLY", _("Monthly")
        QUARTERLY = "QUARTERLY", _("Quarterly")

        @classmethod
        def get_frequence_correspondance(cls, name):
            _map = {
                "DAILY": "D",
                "WEEKLY": "W",
                "MONTHLY": "ME",
                "QUARTERLY": "3M",
            }

            return _map[name]

    name = models.CharField(max_length=255)
    goal = models.PositiveIntegerField(verbose_name=_("Goal"))
    period = DateRangeField(verbose_name=_("Period"))

    evaluated_persons = models.ManyToManyField(
        to="directory.Person", related_name="wbhuman_resources_kpis", verbose_name=_("Evaluated Persons")
    )
    evaluated_intervals = models.CharField(
        max_length=16, choices=Interval.choices, default=Interval.MONTHLY, verbose_name=_("Evaluated Intervals")
    )

    handler = models.CharField(max_length=255)
    additional_data = models.JSONField(default=default_additional_data)
    last_update = models.DateTimeField(
        auto_now=True, verbose_name=_("Last update"), help_text=_("Date of latest change (automatically computed)")
    )
    individual_evaluation = models.BooleanField(default=True, verbose_name=_("Individual Evaluation"))
    is_active = models.BooleanField(default=True, verbose_name=_("Is Active"))

    @staticmethod
    def get_all_handlers() -> Iterator[KPIHandler]:
        for handler_path, name in settings.KPI_HANDLERS:
            *module, handler = handler_path.split(".")
            module_path = ".".join(module)
            yield getattr(importlib.import_module(module_path), handler)()

    @staticmethod
    def get_all_handler_choices() -> Iterator[tuple[str, str]]:
        yield from getattr(settings, "KPI_HANDLERS", [])

    def get_handler(self) -> KPIHandler:
        *module, handler = self.handler.split(".")
        module_path = ".".join(module)
        return getattr(importlib.import_module(module_path), handler)()

    @classmethod
    def is_administrator(cls, user):
        user_groups = user.groups
        user_permission = user.user_permissions
        return (
            user_groups.filter(permissions__codename="administrate_kpi").exists()
            or user_permission.filter(codename="administrate_kpi").exists()
            or user.is_superuser
        )

    def generate_evaluation(self, evaluation_date: date):
        if self.individual_evaluation:
            persons = self.evaluated_persons.all()
        else:
            persons = [None]

        for person in persons:
            if evaluation_date <= date.today() and evaluation_date > self.period.lower:
                Evaluation.objects.update_or_create(
                    evaluation_date=evaluation_date,
                    person=person,
                    kpi=self,
                    defaults={
                        "evaluated_period": [self.period.lower, evaluation_date],
                        "evaluated_score": self.get_handler().evaluate(
                            self, evaluated_person=person, evaluation_date=evaluation_date
                        ),
                    },
                )

    class Meta:
        verbose_name = _("Key Performance Indicator")
        verbose_name_plural = _("Key Performance Indicators")
        permissions = [("administrate_kpi", "Can Administrate KPI")]

    def __str__(self):
        return f"{self.name}"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbhuman_resources:kpi-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbhuman_resources:kpi"


class Evaluation(models.Model):
    """
    Stores the evaluated result of a KPI on the evaluation day for the previous period for 1 person.
    """

    kpi = models.ForeignKey(
        to="wbhuman_resources.KPI", related_name="evaluations", on_delete=models.CASCADE, verbose_name=_("KPI")
    )
    person = models.ForeignKey(
        to="directory.Person",
        related_name="wbhuman_resources_evaluations",
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    evaluated_score = models.IntegerField(null=True, blank=True)
    evaluated_period = DateRangeField()
    evaluation_date = models.DateField()
    last_update = models.DateTimeField(
        auto_now=True, verbose_name=_("Last update"), help_text=_("Date of latest change (automatically computed)")
    )

    class Meta:
        verbose_name = _("Evaluation")
        verbose_name_plural = _("Evaluations")

    def get_rating(self):
        list_date = list(
            pd.date_range(
                start=self.kpi.period.lower,
                end=self.kpi.period.upper,
                freq=KPI.Interval.get_frequence_correspondance(self.kpi.evaluated_intervals),
            )
        )
        list_date = [_date.date() for _date in list_date]
        index = list_date.index(self.evaluation_date)
        nb_intervals = len(list_date)
        goals_expected = (self.kpi.goal * np.ones(nb_intervals) / nb_intervals).cumsum().round(2)
        goal_expected = goals_expected[index]
        rating = (self.evaluated_score / goal_expected * 100).round(2)
        rating = round(rating / 100 * 4)
        return 1 if rating < 1 else rating if rating < 4 else 4



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig

from wbhuman_resources.models import AbsenceRequest, EmployeeHumanResource


class AbsenceRequestEndpointConfig(EndpointViewConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse(f"{self.view.get_model().get_endpoint_basename()}-list", request=self.request)

    def get_instance_endpoint(self, **kwargs):
        if self.instance:
            obj = self.view.get_object()
            if obj.status != AbsenceRequest.Status.DRAFT.name and not EmployeeHumanResource.is_administrator(
                self.request.user
            ):
                return None
        return self.get_list_endpoint()

    def get_create_endpoint(self, **kwargs):
        return self.get_list_endpoint()

    def get_delete_endpoint(self, **kwargs):
        if self.instance:
            obj = self.view.get_object()
            if obj.is_deletable_for_user(self.request.user):
                return self.get_list_endpoint()
        return None


class AbsenceRequestCrossBorderCountryEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbhuman_resources:absencerequesttype-crossbordercountry-list",
            args=[self.view.kwargs["absencerequesttype_id"]],
            request=self.request,
        )


class AbsenceTypeCountEmployeeEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbhuman_resources:employee-absencecount-list",
            args=[self.view.kwargs["employee_id"]],
            request=self.request,
        )


class AbsenceRequestPeriodsAbsenceRequestEndpointConfig(EndpointViewConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbhuman_resources:request-periods-list", args=[self.view.kwargs["request_id"]], request=self.request
        )

    def get_endpoint(self, **kwargs):
        return None


class AbsenceRequestPlannerEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse("wbhuman_resources:absenceplanner-list", args=[], request=self.request)


class AbsenceTablePandasEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse("wbhuman_resources:absencetable-list", args=[], request=self.request)
